from wagtail import blocks
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
import time

from wagtail_editorjs.features import (
    WagtailBlockFeature,
)
from wagtail_editorjs.render import render_editorjs_html
from wagtail_editorjs.registry import (
    EDITOR_JS_FEATURES,
)

from .base import BaseEditorJSTest

class SubBlock(blocks.StructBlock):
    sub_title = blocks.CharBlock()
    sub_text = blocks.CharBlock()

    class Meta:
        allowed_tags = ["h3", "p"]
        allowed_attributes = {
            "h3": ["class"],
            "p": ["class"],
        }

class TestWagtailBlockFeatureBlock(blocks.StructBlock):
    title = blocks.CharBlock()
    subtitle = blocks.CharBlock()
    sub_block = SubBlock()

    class Meta:
        allowed_tags = ["h1", "h2"]
        allowed_attributes = {
            "h1": ["class"],
            "h2": ["class"],
        }

    def render(self, value, context=None):
        return f"<h1 class='test1'>{value['title']}</h1><h2 class='test2'>{value['subtitle']}</h2><h3 class='test3'>{value['sub_block']['sub_title']}</h3><p class='test4'>{value['sub_block']['sub_text']}</p>"


class TestWagtailBlockFeature(BaseEditorJSTest):

    def setUp(self) -> None:
        super().setUp()

        self.block = TestWagtailBlockFeatureBlock()
        self.feature = WagtailBlockFeature(
            "test_feature",
            block=self.block,
        )

        EDITOR_JS_FEATURES.register(
            "test_feature",
            self.feature,
        )

    def test_value_for_form(self):
        test_data = {
            "title": "Test Title",
            "subtitle": "Test Text",
            "sub_block": {
                "sub_title": "Sub Title",
                "sub_text": "Sub Text",
            },
        }

        feature_value = {
            "type": "test_feature",
            "data": {
                "block": test_data,
            }
        }

        editorjs_value = {
            "time": int(time.time()),
            "blocks": [feature_value],
            "version": "0.0.0",
        }

        tdata_copy = test_data.copy()
        copied = editorjs_value.copy()

        data = EDITOR_JS_FEATURES.value_for_form(
            ["test_feature"],
            editorjs_value
        )

        self.assertTrue(isinstance(data, dict))
        self.assertIn("blocks", data)
        self.assertTrue(isinstance(data["blocks"], list))
        self.assertTrue(len(data["blocks"]) == 1, msg=f"Expected 1 block, got {len(data['blocks'])}")

        self.assertDictEqual(
            data,
            copied | {
                "blocks": [
                    {
                        "type": "test_feature",
                        "data": {
                            "block": tdata_copy,
                        }
                    }
                ]
            }
        )
        

    def test_wagtail_block_feature(self):
        test_data = {
            "title": "Test Title",
            "subtitle": "Test Text",
            "sub_block": {
                "sub_title": "Sub Title",
                "sub_text": "Sub Text",
            },
        }

        feature_value = {
            "type": "test_feature",
            "data": {
                "block": test_data,
            }
        }

        editorjs_value = {
            "time": int(time.time()),
            "blocks": [feature_value],
            "version": "0.0.0",
        }

        html = render_editorjs_html(features=["test_feature"], data=editorjs_value)
        feature_html = str(self.feature.render_block_data(feature_value))
        
        
        self.assertHTMLEqual(
            html,
            render_to_string(
                "wagtail_editorjs/rich_text.html",
                {"html": mark_safe(feature_html)}
            ),
        )

        self.assertInHTML(
            "<h1 class='test1'>Test Title</h1>",
            feature_html,
        )

        self.assertInHTML(
            "<h2 class='test2'>Test Text</h2>",
            feature_html,
        )

        self.assertInHTML(
            "<h3 class='test3'>Sub Title</h3>",
            feature_html,
        )

        self.assertInHTML(
            "<p class='test4'>Sub Text</p>",
            feature_html,
        )




from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from wagtail.telepath import register
from wagtail.blocks.field_block import (
    FieldBlock,
    FieldBlockAdapter,
)

from .render import render_editorjs_html
from .forms import (
    EditorJSFormField,
    get_features,
)


class EditorJSBlock(FieldBlock):
    """
        A Wagtail block which can be used to add the EditorJS
        widget into any streamfield or structblock.
    """
    class Meta:
        icon = 'draft'
        label_format = _('EditorJS Block')

    def __init__(self, features: list[str] = None, tools_config: dict = None, **kwargs):
        self._features = features
        self.tools_config = tools_config or {}
        super().__init__(**kwargs)

    @cached_property
    def field(self):
        return EditorJSFormField(
            features=self.features,
            tools_config=self.tools_config,
            label=getattr(self.meta, 'label', None),
            required=getattr(self.meta, 'required', True),
            help_text=getattr(self.meta, 'help_text', ''),
        )
    
    @property
    def features(self):
        return get_features(self._features)

    def render(self, value, context=None):
        """
            Render the block value into HTML.
            This is so the block can be automatically included with `{% include_block my_editor_js_block %}`.
        """
        return render_editorjs_html(self.features, value, context)


from django.forms import Media


class EditorJSBlockAdapter(FieldBlockAdapter):
    js_constructor = "wagtail_editorjs.blocks.EditorJSBlock"

    @cached_property
    def media(self):
        m = super().media
        return Media(
            js= m._js + [
                "wagtail_editorjs/js/editorjs-block.js",
            ],
            css=m._css,
        )

register(EditorJSBlockAdapter(), EditorJSBlock)



from typing import TYPE_CHECKING
from django.urls import path
from django.http import HttpRequest, HttpResponseNotAllowed
from wagtail import hooks

if TYPE_CHECKING:
    from ..feature_registry import EditorJSFeatures


class FeatureViewMixin:

    def handler(self, request: HttpRequest, *args, **kwargs):
        method = request.method.lower()
        if not hasattr(self, f"handle_{method}"):
            return self.method_not_allowed(request)
        
        view_func = getattr(self, f"handle_{method}")
        return view_func(request, *args, **kwargs)

    def method_not_allowed(self, request: HttpRequest):
        methods = ["get", "post", "put", "patch", "delete"]
        methods = [m for m in methods if hasattr(self, f"handle_{m}")]
        return HttpResponseNotAllowed(methods)

    def get_urlpatterns(self):
        return [
            path(
                f"{self.tool_name}/",
                self.handler,
                name=self.tool_name,
            )
        ]

    def on_register(self, registry: "EditorJSFeatures"):
        super().on_register(registry)
        @hooks.register("register_editorjs_urls")
        def register_admin_urls():
            return self.get_urlpatterns()




from django.utils.functional import cached_property
from django.utils.safestring import mark_safe
from .forms import EditorJSFormField, EditorJSWidget



class DjangoEditorJSFormField(EditorJSFormField):
    @cached_property
    def widget(self):
        return EditorJSDjangoWidget(
            features=self.features,
            tools_config=self.tools_config,
        )



class EditorJSDjangoWidget(EditorJSWidget):
    """
        Taken from wagtail.
        This class is deprecated in wagtail 7.0
        This might still be useful if we have wagtail installed
        but would like to use the editorjs widget in a non-wagtailadmin form.
    """
    def render_html(self, name, value, attrs):
        """Render the HTML (non-JS) portion of the field markup"""
        return super().render(name, value, attrs)

    def render(self, name, value, attrs=None, renderer=None):
        # no point trying to come up with sensible semantics for when 'id' is missing from attrs,
        # so let's make sure it fails early in the process
        try:
            id_ = attrs["id"]
        except (KeyError, TypeError):
            raise TypeError(
                "WidgetWithScript cannot be rendered without an 'id' attribute"
            )

        value_data = self.get_value_data(value)
        widget_html = self.render_html(name, value_data, attrs)

        js = self.render_js_init(id_, name, value_data)
        out = f"{widget_html}<script type=\"text/javascript\">{js}</script>"
        return mark_safe(out)

    def render_js_init(self, id_, name, value):
        # Adapted from editorjs-widget-controller.js
        return """
let editorJSWidget__wrapper = document.querySelector(`#${id}-wagtail-editorjs-widget-wrapper`);
let editorJSWidget__configElem = editorJSWidget__wrapper.querySelector(`#wagtail-editorjs-config`);
let editorJSWidget__config = JSON.parse(editorJSWidget__configElem.textContent);
let editorJSWidget__keys = Object.keys(editorJSWidget__config.tools);
for (let i = 0; i < editorJSWidget__keys.length; i++) {
    const key = editorJSWidget__keys[i];
    const toolConfig = editorJSWidget__config.tools[key];
    const toolClass = window[toolConfig.class];
    toolConfig.class = toolClass;
    editorJSWidget__config.tools[key] = toolConfig;
}
let editorJSWidget__element = document.querySelector(`#${id}`);
new window.EditorJSWidget(
    editorJSWidget__wrapper,
    editorJSWidget__element,
    editorJSWidget__config,
);
""" % {
            "id": id_,
        }




from wagtail_editorjs.render import render_editorjs_html
from wagtail_editorjs.registry import EditorJSValue
from wagtail_editorjs.fields import EditorJSField
from wagtail import hooks



@hooks.register("wagtail_fedit.register_type_renderer")
def register_renderers(renderer_map):

    # This is a custom renderer for RichText fields.
    # It will render the RichText field as a RichText block.
    renderer_map[EditorJSValue] = lambda request, context, instance, value: render_editorjs_html(
        value.features,
        value,
        context=context,
    )

@hooks.register("wagtail_fedit.field_editor_size")
def field_editor_size(model_instance, model_field):
    if isinstance(model_field, EditorJSField):
        return "full"
    return None



from typing import Any, Union, TYPE_CHECKING
import datetime

if TYPE_CHECKING:
    from .features import EditorJSFeature


class EditorJSBlock(dict):
    def __init__(self, data: dict, features: list[str]):
        self.features = features
        super().__init__(data)

    @property
    def id(self):
        return self.get("id")
    
    @property
    def type(self):
        return self.get("type")
    
    @property
    def data(self):
        return self.get("data", {})
    
    @property
    def tunes(self):
        return self.get("tunes", {})


class EditorJSValue(dict):
    def __init__(self, data: dict, features: dict[str, "EditorJSFeature"]):
        self._features = features
        super().__init__(data)

    @property
    def blocks(self) -> list["EditorJSBlock"]:
        return self["blocks"]
    
    @property
    def time(self):
        time = self.get("time")
        if time is None:
            return None
        return datetime.datetime.fromtimestamp(time)
    
    @property
    def version(self):
        return self.get("version")
    
    @property
    def features(self) -> list["EditorJSFeature"]:
        return self._features
    
    def __getitem__(self, __key: Any) -> Any:
        if isinstance(__key, int):
            return self.blocks[__key]
        
        return super().__getitem__(__key)

    def get_blocks_by_name(self, name: str) -> list["EditorJSBlock"]:
        if name not in self._features:
            return []
        
        return [
            block
            for block in self.blocks
            if block.get('type') == name
        ]
    
    def get_block_by_id(self, id: str) -> "EditorJSBlock":
        for block in self.blocks:
            if block.get("id") == id:
                return block
        return None
    
    def get_range(self, start: int, end: int) -> list["EditorJSBlock"]:
        return self["blocks"][start:end]
    
    def set_range(self, start: int, end: int, blocks: list["EditorJSBlock"]):
        b = self["blocks"]
        if start < 0:
            start = 0
        
        if end > len(b):
            end = len(b)
        
        if start > end:
            start = end
        
        if start == end:
            return
        
        b[start:end] = list(
            map(self._verify_block, blocks),
        )
        self["blocks"] = b

    def insert(self, index: int, block: "EditorJSBlock"):
        block = self._verify_block(block)        
        self.blocks.insert(index, block)

    def append(self, block: "EditorJSBlock"):
        block = self._verify_block(block)
        self.blocks.append(block)

    def _verify_block(self, block: Union["EditorJSBlock", dict]):
        if block.get("type") not in self._features:
            raise ValueError(f"Unknown feature: {block.get('type')}")
        
        if block.get("id") is None:
            raise ValueError("Block ID not set")
        
        if not isinstance(block, EditorJSBlock)\
            and isinstance(block, dict):
            
            block = self._features[block["type"]].create_block(
                list(self._features.keys()),
                block
            )

        return block



from typing import Any, Union, TYPE_CHECKING
from django.urls import reverse_lazy
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe
from ..value import (
    EditorJSBlock,
)

if TYPE_CHECKING:
    from ..feature_registry import EditorJSFeatures


class TemplateNotSpecifiedError(Exception):
    pass


class PageChooserURLsMixin:
    def get_config(self, context: dict[str, Any] = None) -> dict:
        config = super().get_config(context)
        config.setdefault("config", {})
        config["config"]["chooserUrls"] = {
            "pageChooser": reverse_lazy(
                "wagtailadmin_choose_page",
            ),
            "externalLinkChooser": reverse_lazy(
                "wagtailadmin_choose_page_external_link",
            ),
            "emailLinkChooser": reverse_lazy(
                "wagtailadmin_choose_page_email_link",
            ),
            "phoneLinkChooser": reverse_lazy(
                "wagtailadmin_choose_page_phone_link",
            ),
            "anchorLinkChooser": reverse_lazy(
                "wagtailadmin_choose_page_anchor_link",
            ),
        },
        return config


class BaseEditorJSFeature:
    allowed_tags: list[str] = None
    allowed_attributes: dict[str, list[str]] = None
    klass: str = None
    js: list[str] = None
    css: list[str] = None
    frontend_css: list[str] = []
    frontend_js: list[str] = []
    # cleaner_funcs: dict[str, dict[str, Callable[[str], bool]]] = {}
    
    def __init__(self,
            tool_name: str,
            klass: str = None,
            js: Union[str, list[str]] = None,
            css: Union[str, list[str]] = None,
            include_template: str = None,
            config: dict = None,
            weight: int = 0, # Weight is used to manage which features.js files are loaded first.
            allowed_tags: list[str] = None,
            allowed_attributes: dict[str, list[str]] = None,
            **kwargs
        ):

        if not klass and not self.klass:
            raise ValueError("klass must be provided")
        
        if not klass:
            klass = self.klass
        
        self.tool_name = tool_name
        self.klass = klass
        self.config = config or dict()
        self.kwargs = kwargs
        self.weight = weight
        self.include_template = include_template

        self.init_static(
            css, js,
        )

        self.init_attrs(
            allowed_tags, allowed_attributes,
        )

    def value_for_form(self, value: dict) -> dict:
        """
            Prepare the value for the feature.
            This is useful for when you need to modify the data
            before it is passed to the frontend.
        """
        return value

    def init_static(self, css, js):
        css = css or []
        js = js or []

        if self.css:
            css.extend(self.css)

        if self.js:
            js.extend(self.js)
            
        self.js = js
        self.css = css

    def init_attrs(self, allowed_tags, allowed_attributes):
        if not isinstance(allowed_tags, (list, tuple, set)) and allowed_tags is not None:
            raise ValueError("allowed_tags must be a list, tuple or set")
        
        if not isinstance(allowed_attributes, dict) and allowed_attributes is not None:
            raise ValueError("allowed_attributes must be a dict")

        allowed_tags = allowed_tags or []
        allowed_attributes = allowed_attributes or dict()
       
        if self.allowed_tags:
            allowed_tags.extend(self.allowed_tags)

        if self.allowed_attributes:

            if isinstance(self.allowed_attributes, dict):
                for key, value in self.allowed_attributes.items():
                    allowed_attributes[key] = set(allowed_attributes.get(key, []) + value)
            elif isinstance(self.allowed_attributes, list):
                if not self.allowed_tags:
                    raise ValueError("Allowed attributes is specified as a list without allowed tags; provide allowed tags.")
                
                for tag in self.allowed_tags:
                    allowed_attributes[tag] = set(allowed_attributes.get(tag, []) + self.allowed_attributes)
            else:
                raise ValueError("Invalid allowed attributes type, self.allowed_attributes must be dict or list")

        self.allowed_tags = set(allowed_tags)
        self.allowed_attributes = allowed_attributes

    def on_register(self, registry: "EditorJSFeatures"):
        """
            Called when the feature is registered.
            This can be used to say; register URLs
            required for this feature to work.
        """
        pass

    def __repr__(self):
        return f"<EditorJSFeature \"{self.tool_name}\">"
    
    @classmethod
    def get_test_data(cls):
        """
            Returns test data for the feature.
            This is so it can easily be integrated with our tests.
            Any extra testing you deem necessary should be done in
            separate tests.
        """
        return []

    def get_template_context(self, context: dict[str, Any] = None) -> dict:
        """
            Returns the context for the template. (if any template is provided)
        """
        return context
    
    def render_template(self, context: dict[str, Any] = None):
        """
            Renders a template inside of the widget container.
            This is useful for things like the image feature.
        """
        if self.include_template and hasattr(self.include_template, "render"):
            return self.include_template.render(self.get_template_context(context))
        elif self.include_template:
            context = self.get_template_context(context)
            rendered = render_to_string(self.include_template, context)
            return mark_safe(rendered)
        raise TemplateNotSpecifiedError("Template not specified for this feature")
    
    def get_config(self, context: dict[str, Any] = None) -> dict:
        """
            Returns the config for the feature.
            This is what will be passed to javascript.
            The `class` is extracted from window[`class`].
        """
        config = {
            "class": self.klass,
        }
        
        if self.config:
            config["config"] = self.config

        if self.kwargs:
            config.update(self.kwargs)

        return config

    def get_js(self):
        """
            Return any javascript files required for this feature to work.
        """
        if not self.js:
            return []
            
        if isinstance(self.js, str):
            return [self.js]
        return self.js
    
    def get_css(self):
        """
            Return any css files required for this feature to work.
        """
        if not self.css:
            return []
            
        if isinstance(self.css, str):
            return [self.css]
        return self.css
    
    def get_frontend_css(self):
        """
            Returns the css files required for the frontend.
        """
        return self.frontend_css
    
    def get_frontend_js(self):
        """
            Returns the js files required for the frontend.
        """
        return self.frontend_js

    def validate(self, data: Any):
        """
            Validate any data coming from editorJS
            for completeness and correctness.
        """
        pass

    def create_block(self, tools: list[str], data: dict) -> EditorJSBlock:
        """
            Create a block from the data.
            This block is the value that the developer will work with.
            It is a subclass of `dict`.
        """
        return EditorJSBlock(data, tools)
